/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.assertlevel;

/**
 * <h2>Rationale</h2>
 * Although it may be tempting to "group" asserts into well-defined categories and turn each category on and off, 
 * providing such flexibility seems to be unneded since it brings a lot of complexity and thus takes more than it gives.
 * Therefore, threat levels are made very similar to log4j logging levels. If you turn off threat level 
 * {@link ThreatLevel#INTERNAL_INVARIANT_CHECK} then levels {@value ThreatLevel#INTERNAL_INVARIANT_CHECK} and
 * {@value ThreatLevel#PARANOID} also will be turned off.
 * So, in production, you have several options to run a program.
 * <ol>
 * 	<li>turn off assertions (no -ea switch), cutting off this facility effectively</li>
 *  <li>specify level={@link ThreatLevel#INPUT_PARAMETERS_CHECK} to allow additional input parameters checking</li>
 *  <li>specify level={@link ThreatLevel#OUTPUT_PARAMETERS_CHECK} to allow additional input and output parameters 
 *      checking</li>
 *  <li>specify level={@link ThreatLevel#INTERNAL_INVARIANT_CHECK} to allow additional input/output parameters 
 *      checking and also checking internal invariants</li>
 *  <li>specify level={@link ThreatLevel#PARANOID} to allow additional input/output parameters checking,
 *      internal invariants checking and any paranoid-level checks developer thought posssible</li>
 *  <li>define your own custom threat level</li>
 * </ol>
 * 
 * <h2>Usage patterns</h2>
 * Although using these levels and assigning checks to levels is up to the developer, the following "best practices"
 * are proposed:
 * <ul>
 * 	<li>Use {@link ThreatLevel#STANDARD} level by default, for example initially.</li>
 *  <li>Try to check your input parameters using regular java code and throwing errors (without this facility)</li>
 *  <li>Use {@link ThreatLevel#INPUT_PARAMETERS_CHECK} when you need to perform additional expensive checks on input
 *      parameters</li>
 *  <li>Use {@link ThreatLevel#OUTPUT_PARAMETERS_CHECK} when you need to perform additional expensive checks on result
 *      returned from method</li>
 *  <li>Use {@link ThreatLevel#INTERNAL_INVARIANT_CHECK} when you need to perform additional expensive checks internal
 *      object state</li>
 *  <li>Use PARANOID for paranoid checks</li>
 * </ul>
 * 
 * @author borisman
 *
 */
public enum ThreatLevel {
	/** 
	 * Turn off assertlevel completely. Useful when you want to turn this facility off, but do not want to turn off all
	 * assertions by ommiting -ea switch
	 */
	OFF(0),
	
	/**
	 * All assertions you have not categorized yet. Use this as default in production 
	 */
	STANDARD(10000),
	
	/**
	 * Additional expensive input parameters checking. 
	 * NOTE!!! You generally should check your input parameters without asserts, but when check has already been 
	 * performed somewhere and you just want to double-check, use this.
	 * 
	 * Example 1 (trivial): 
	 * <pre>
	 *   public void setName(String name) {
	 *     if (name.length()==0) { //ALWAYS DO THIS!
	 *       throw new IllegalArgumentException("name cannot be empty");
	 *     }
	 *     
	 *     this.setNameImpl(name);
	 *   }
	 *   
	 *   private void setNameImpl(String name) {
	 *     // At this point, name should already been checked by setName, but we want to be sure
	 *     assert AL.input() || name.length()>0 : "Name cannot be empty"; 
	 *   }
	 * </pre>
	 * 
	 * Example 2 (more realistic):
	 * You have an API with some public method:
	 * <pre>
	 *   public void setName(String name) {
	 *     if (name.length()==0) { //ALWAYS DO THIS!
	 *       throw new IllegalArgumentException("name cannot be empty");
	 *     }
	 *     ...
	 *   }
	 * </pre>
	 * Now, several layers deep down the calling stack you have:
	 * <pre>
	 *   private void doSomethingWithName(String name) {
	 *     // At this point, name should already been checked by public API methods, but we want to be sure
	 *     assert AL.input() || name.length()>0 : "Name cannot be empty"; 
	 *     ...
	 *   }
	 * </pre>
	 *  
	 */
	INPUT_PARAMETERS_CHECK(20000),
	
	/**
	 * Output parameters checking.
	 * Example: 
	 * <pre>
	 *   public String computeName() {
	 *     ...
	 *     assert AL.output() || res.length()>0 : "Invalid result, cannot be empty";
	 *     return res;
	 *   }
	 * </pre>
	 */
	OUTPUT_PARAMETERS_CHECK(30000),
	
	/**
	 * Internal invariants checking.
	 * Example: 
	 * <pre>
	 *   public void doSomething() {
	 *     ...
	 *     assert AL.invariant() || this.list.size()==0 : "list should not be modified!";
	 *     ...
	 *   }
	 * </pre>
	 */
	INTERNAL_INVARIANT_CHECK(40000),
	
	/**
	 * Paranoid checks (default value in configuration).  
	 */
	PARANOID(50000);
	
	public final int code;
	
	private ThreatLevel(final int code) {
		this.code = code;
	}

	public static ThreatLevel fromName(final String expr) {
		for (final ThreatLevel level : ThreatLevel.values()) {
			if (level.name().equals(expr)) return level;
		}
		return null;
	}
	
}
