/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.collections.fields;

import java.io.Externalizable;
import java.io.IOException;
import java.io.ObjectInput;
import java.io.ObjectOutput;
import java.util.Iterator;

import org.builtu.collections.adapters.ArrayIterator;
import org.builtu.util.commutil.Util;


/**
 * Very simple class that provides lightweight name to value mapping.
 * Recommended for small collections with fixed number of items. 
 *
 * @author borisman
 */
public class Fields implements Cloneable, Externalizable {

	@Override public boolean equals(final Object rval) {
		if (this == rval) return true;
		if (! (rval instanceof Fields)) return false;
		final Fields r = (Fields) rval;
		if (names.length!=r.names.length) return false;
		for (int i=0; i<names.length; i++) {
			if (names[i]!=r.names[i]) return false;
			final Object o1 = vals[i],
			             o2 = r.vals[i];
			if (o1 == null && o2 != null) return false;
			if (o1 != null && !o1.equals(o2)) return false;
		}
		return true;
	}

	private static final long serialVersionUID = 1L;
	/*
     * -------------------------------------------------------------------------
     * Attributes
     * -------------------------------------------------------------------------
     */
    /** This holds names of the fields*/
    private String[] names;
    /** This holds values of the fields. This is and Object[] array since
     * {@link Fields#Fields(String[])} constructor can't create V[] arrays.*/
    private String[] vals;
    /** length == names.length == values.length */
    public int length;

    /*
     * -------------------------------------------------------------------------
     * Constructors
     * -------------------------------------------------------------------------
     */
    /**
     * For serialization only!!!
     */
    public Fields() {
    }
    /**
     * Create fields from names supplied as array of strings. All values will be
     * initialized to null
     *
     * @throws IllegalArgumentException when names contains duplicate strings
     */
    public Fields(final String[] names) throws IllegalArgumentException {
        this(names, new String[names.length]);
    }
    /**
     * Create fields with default values. Contents of the Object[] values will
     * be copied into internal values array.
     * @throws IllegalArgumentException when names.length != values.length or
     *         names contains duplicate strings
     */
    public Fields(final String[] names, final String[] values)
            throws IllegalArgumentException {
        if (names.length == 0)
        	throw new IllegalArgumentException("No point in 0-length fields");
        if (names.length != values.length)
            throw new IllegalArgumentException("names.length != values.length");

        this.names = new String[names.length];
        for (int i = 0; i < this.names.length; i++)
            this.names[i] = names[i].intern();
        this.length = names.length;
        this.vals = values;
        checkDuplicateNames();
    }

    /*
     * -------------------------------------------------------------------------
     * Interface
     * -------------------------------------------------------------------------
     */
    /**
     * @return name of the field by it's index
     */
    public String name(final int idx) {
        return names[idx];
    }

    public Iterator<String> names () {
        return new ArrayIterator<String>(names);
    }

    /**
     * @return true if has field with given name
     */
    public boolean containsName(String name) {
        name = name.intern();
        for (String n : names)
            if (n == name)
                return true;
        return false;
    }

    /**
     * Return field value by it's index. Since set methods always insert values
     * of the "right" class, we can suppress cast warning exception here
     */
    public String get(final int idx) {
        return vals[idx];
    }
    /**
     * Return field value by field name. Since set methods always insert values
     * of the "right" class, we can suppress cast warning exception here
     *
     * @throws IndexOutOfBoundsException when no field exists with the name
     *                                   given
     */
    public String get(String name) throws IndexOutOfBoundsException {
        name = name.intern();
        for (int i = 0; i < names.length; i++)
            if (names[i] == name)
                return vals[i];
        throw new IndexOutOfBoundsException(name);
    }

    /**
     * Set field value by field index
     */
    public void set(final int idx, final String value) {
        vals[idx] = value;
    }

    /**
     * Set field value by field name
     * @throws IndexOutOfBoundsException when no field exists with the name
     *                                   given
     */
    public void set(String name, final String value)
            throws IndexOutOfBoundsException
    {
        name = name.intern();
        for (int i = 0; i < names.length; i++)
            if (names[i] == name) {
                vals[i] = value;
                return;
            }
        throw new IndexOutOfBoundsException(name);
    }

    /**
     * Set all field values to null
     */
    public void clear() {
       clear(null);
    }

    /**
     * Set all field values to specified value
     */
    public void clear(final String fill) {
        for (int i = 0; i < vals.length; i++)
            vals[i] = fill;
    }
    
    /*
     * -------------------------------------------------------------------------
     * Implementation
     * -------------------------------------------------------------------------
     */
    private void checkDuplicateNames() {
        for (int i = 0; i < length - 1; i++)
            for (int j = i + 1; j < length; j++)
                if (names[i] == names[j])
                    throw new IllegalArgumentException("Duplicate name "+ names[i] + " detected");
    }

    public Fields clone() throws CloneNotSupportedException {
        Fields cloned = (Fields) super.clone();
        cloned.names = this.names.clone();
        cloned.vals = this.vals.clone();
        return cloned;
    }

    public int indexOfField(final String fieldName) {
		return Util.indexOf(names, fieldName);
	}
	
    @Override public void readExternal(final ObjectInput in) throws IOException, ClassNotFoundException {
    	names = (String[]) in.readObject();
    	vals = (String[]) in.readObject();
    	length = names.length;
    	assert names.length == vals.length;
    	
    	for (int i=0; i<length; i++) names[i] = names[i].intern();
	}
	
    @Override public void writeExternal(final ObjectOutput out) throws IOException {
		out.writeObject(names);
		out.writeObject(vals);
	}
}