/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.testing.webmock;

import java.io.IOException;
import java.io.OutputStream;
import java.io.UnsupportedEncodingException;
import java.net.URL;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.helpers.DefaultValidationEventHandler;

import org.builtu.exceptions.FatalError;
import org.builtu.util.commutil.StringUtil;
import org.builtu.util.commutil.Util;
import org.builtu.webmock.XmlWebAlias;
import org.builtu.webmock.XmlWebAliases;
import org.builtu.webmock.XmlWebParam;
import org.builtu.webmock.XmlWebRequest;
import org.builtu.webmock.XmlWebResponse;
import org.mortbay.jetty.Server;
import org.mortbay.jetty.handler.AbstractHandler;


public class WebMock {
	private final Map<String, Map<Request, Result>> results = new HashMap<String, Map<Request, Result>>();
	private final Map<String, String> pathAliases = new HashMap<String, String>();
	public final List<Filter> filters = new ArrayList<Filter>();
	private int port; 
	
	private Server server;
	
	/**
	 * Add alias for path. Aliased paths will be treated same as regular paths, so when request goes to aliased
	 * path, it will be redirected to it's original path before any filters apply.
	 * 
	 * Very useful if you want to treat paths "/" and "/index.html" like same paths 
	 * @param path1
	 * @param path2
	 */
	public void alias(final String path1, final String path2) {
		Request.checkPath(path1);
		Request.checkPath(path2);
		
		pathAliases.put(path1, path2);
		pathAliases.put(path2, path1);
	}

	public void addMapping(final Request request, final Result result) {
		Util.put(results, request.getPath(), request, result);
	}

	
	public void handle(Request request, final HttpServletResponse response) throws IOException {
		for (final Filter f : filters) {
			if (f.filter(request, this, response)) return;
		}
		
		Result res = Util.get(results, request.getPath(), request);
		if (res == null) { //try aliased path
			final String aliased = pathAliases.get(request.getPath());
			if (aliased != null) {
				request.setPath(aliased);
				res = Util.get(results, request.getPath(), request);
			}
		}
		
		if (res==null) {
			response.sendError(404);
			return;
		}
		String contentType = res.contentType;
		if (!StringUtil.isEmpty(res.encoding)) contentType += "; charset="+res.encoding;
		response.setContentType(contentType);
		response.setContentLength(res.data.length);
		final OutputStream out = response.getOutputStream();
		try {
			out.write(res.data);
		} finally {
			out.close();
		}
	}
	
	private final class MockHandler extends AbstractHandler {
		
		private final WebMock webMock;
		
		public MockHandler(final WebMock webMock) {
			this.webMock = webMock;
		}

		@Override public void handle(
				final String target, 
				final HttpServletRequest request,
				final HttpServletResponse response, 
				final int dispatch
		) throws IOException, ServletException {
			final Request req = new Request();
			req.setPath(target);
			req.parseHttpRequest(request);
			webMock.handle(req, response);
		}
	}
	
	public void start(final int port) throws Exception {
		if (server != null) throw new FatalError("Already started");
		final Server server = new Server(port);
		final MockHandler handler = new MockHandler(this);
		server.setHandler(handler);
		server.start();
		this.server = server;
		this.port = port;
	}

	public void stop() throws Exception {
		if (server == null) throw new FatalError("Not started");
		server.stop();
		server = null;
	}
	
	public int getPort() {
		if (server == null) throw new FatalError("Not started");
		return port;
	}

	public void readConfig(final URL location) {
		final org.builtu.webmock.Webmock webmock;
		try {
			final JAXBContext context = JAXBContext.newInstance("org.builtu.webmock");
			final Unmarshaller unmarshaller = context.createUnmarshaller();
			unmarshaller.setEventHandler(new DefaultValidationEventHandler());
			webmock = (org.builtu.webmock.Webmock) unmarshaller.unmarshal(location);
		} catch (final JAXBException e) {
			throw new FatalError(e);
		}
		
		readWebRequests(webmock.getRequests().getRequest());
		readPathAliases(webmock.getAliases());
	}

	private void readPathAliases(final XmlWebAliases pathAliases) {
		if (pathAliases == null) return;
		for (final XmlWebAlias alias : pathAliases.getAlias()) {
			alias(alias.getPath1(), alias.getPath2());
		}
	}
	
	private final static class RequestResultTuple {
		public Request request;
		public Object result;
		
		public RequestResultTuple(final Request request, final Object result) {
			this.request = request;
			this.result = result;
		}
	}

	private void readWebRequests(final List<XmlWebRequest> requests) {
		if (requests == null) return;
		final List<RequestResultTuple> results = new ArrayList<RequestResultTuple>(requests.size());
		final Map<String, Result> resultsByName = new HashMap<String, Result>(requests.size());
		
		for (final XmlWebRequest request : requests) {
			final Request r = new Request();
			final Map<String, List<String>> params = new HashMap<String, List<String>>();
			r.setPath(parseUrl(request.getUrl(), params));
			parseParamsFromTags(request.getParam(), params);
			setParams(r, params);
			
			final XmlWebResponse response = request.getResponse();
			if (!StringUtil.isEmpty(response.getRef())) {
				results.add(new RequestResultTuple(r, response.getRef()));
			} else {
				final Result res = new Result();
				res.contentType = response.getContentType();
				res.encoding = response.getEncoding();
				try {
					res.data = response.getBody().getBytes(res.encoding);
				} catch (final UnsupportedEncodingException e) {
					throw new FatalError(e);
				}
				final String name = response.getName();
				if (!StringUtil.isEmpty(name)) {
					if (resultsByName.put(name, res)!=null) {
						throw new FatalError("There are at least two responses named "+name);
					}
				}
				
				results.add(new RequestResultTuple(r, res));
			}
		}

		for (final RequestResultTuple t : results) {
			final Object val = t.result;
			final Result res;
			if (val instanceof Result) {
				res = (Result) val;
			} else {
				res = resultsByName.get((String)val);
				if (res == null) throw new FatalError("No result found: "+val);
			}
			
			final Result old = Util.put(this.results, t.request.getPath(), t.request, res);
			
			if (old != null) {
				throw new FatalError("There are two results mapped to "+t.request+", first="+old+
						             ", second="+res);
			}
		}
	}

	private void setParams(final Request r, final Map<String, List<String>> params) {
		for (final Map.Entry<String, List<String>> e : params.entrySet()) {
			final String [] arr = new String[e.getValue().size()];
			e.getValue().toArray(arr);
			r.set(e.getKey(), arr);
		}
	}

	private String parseUrl(String url, final Map<String, List<String>> params) {
		url = skipHost(url);
		final int dynPart = url.indexOf('?');
		if (dynPart<0) {
			return url;
		} else {
			final String path = url.substring(0, dynPart);
			final String dyn = url.substring(dynPart+1);
			for (final String nv : StringUtil.split_list(dyn, '&')) {
				final int eq = nv.indexOf('=');
				final String name;
				final String value; 
				if (eq<0) {
					name = nv;
					value = null;
				} else {
					name = Util.urldecode(nv.substring(0, eq));
					value = Util.urldecode(nv.substring(eq+1)); 
				}
				
				List<String> vals = params.get(name);
				if (vals == null) {
					vals = new ArrayList<String>();
					params.put(name, vals);
				}
				if (value != null) vals.add(value);
			}
			return path;
		}
	}

	private String skipHost(String url) {
		url = url.trim();
		int idx = url.indexOf('/');
		if (idx == 0) return url;
		if (url.charAt(idx-1)==':' && idx<url.length()-1 && url.charAt(idx+1)=='/') {
			idx = url.indexOf('/', idx+2);
		}
		return url.substring(idx);
	}

	private void parseParamsFromTags(
			final List<XmlWebParam> params, 
			final Map<String, List<String>> map
	) {
		for (final XmlWebParam param : params) {
			final String name = param.getName(); //;
			List<String> values = map.get(name);
			if (values == null) {
				values = new ArrayList<String>();
				map.put(name, values);
			}
			if (param.getValue()!=null) values.addAll(param.getValue());
		}
	}

}