/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.commutil;

import java.util.ArrayList;
import java.util.List;
import java.util.Random;

import org.builtu.testing.BasicTestCase;
import org.builtu.util.commutil.Util;
import org.junit.Test;


public class TestUtilRandom extends BasicTestCase {
	
	private List<Integer> makeList(final int size) {
		final List<Integer> res = new ArrayList<Integer>(size);
		for (int i=0; i<size; i++) res.add(i);
		return res;
	}
	
	@Test(expected=IllegalArgumentException.class) public void badLength() {
		final List<Integer> source = new ArrayList<Integer>();
		final Integer[] dest = new Integer[2];
		Util.randomSample(super.random, source, dest);
	}

	@Test public void test10() {
		doTest(1, 0);
	}
	
	@Test public void test11() {
		doTest(1, 1);
	}

	@Test public void test21() {
		doTest(2, 1);
	}

	@Test public void test22() {
		doTest(2, 2);
	}

	@Test public void test31() {
		doTest(3, 1);
	}

	@Test public void test32() {
		doTest(3, 2);
	}

	@Test public void test33() {
		doTest(3, 3);
	}
	
	@Test public void testIntegerOpt() {
		final int N = rnd(30);
		final int K = rnd(N);
		doTest(N, K);
	}
	
	@Test public void testRandomSubList() {
		final int N = rnd(100);
		final int K = rnd(N+1);
		doTest(N, K);
	}
	
	@Test public void testRandomInclusion() {
		final int N = 200;
		final int K = 90;
		doTest(N, K);
	}
	
	@Test public void testRandomExclusion() {
		final int N = 200;
		final int K = 190;
		doTest(N, K);
	}
	
	private void doTest(final int N, final int K) {
		final List<Integer> list = makeList(N);
		final Integer[] random = new Integer[K]; 
		Util.randomSample(super.random, list, random);
		checkValid(list, random);
		
		checkDistributionQuality(super.random, list, random);
	}

	private void checkDistributionQuality(final Random rnd, final List<Integer> list, final Integer[] random) {
		final int RUNS = 10000;
		final int [] frequences = new int [list.size()];
		final int [] randomGeneratorTest = new int [list.size()];
		for (int i=0; i<RUNS; i++) {
			Util.randomSample(super.random, list, random);
			for (final int value : random) {
				frequences[value]++;
				final int idx = rnd.nextInt(randomGeneratorTest.length);
				randomGeneratorTest[idx]++;
			}
		}
		
		final int randomGeneratorDeviationPercent = calculateDeviationPercent(randomGeneratorTest);
		final int samplerDeviationPercent = calculateDeviationPercent(frequences);
		System.out.println("generator error="+(randomGeneratorDeviationPercent/100)+"%, " +
				           "sampler error="+(samplerDeviationPercent/100)+"%");
		
		final boolean deviationIsTooSmall = samplerDeviationPercent<200;
		final boolean deviationSignificantlyWorseThanGenerators = samplerDeviationPercent<=2*randomGeneratorDeviationPercent;
		final boolean unacceptable = !deviationIsTooSmall && deviationSignificantlyWorseThanGenerators; 
		
		assertFalse("Sampler deviation is unacceptable ("+
				   "generator="+(randomGeneratorDeviationPercent/100)+"%, sampler="+samplerDeviationPercent/100+"%)",
				   unacceptable); 
	}

	private int calculateDeviationPercent(final int[] frequences) {
		int total = 0;
		for (final int f : frequences) total+=f;
		if (total == 0) return 0;
		
		final int average = total / frequences.length;
		int maxdev = 0;
		for (final int f : frequences) {
			final int deviation = Math.abs(f - average);
			if (deviation > maxdev) maxdev = deviation;
		}
		
		return 10000 * maxdev / total;
	}

	private void checkValid(final List<Integer> list, final Integer[] random) {
		assertTrue(ordered(random));
		for (int i : random) {
			assertTrue(list.contains(i));
		}
	}

	private boolean ordered(final Integer[] list) {
		if (list.length==0) return true; 
		int prev = list[0];
		for (int i=1; i<list.length; i++) {
			final int current = list[i];
			if (current<=prev) return false;
		}
		return true;
	}
}
