/**
	This file is part of BUILTU.

    BUILTU is free software: you can redistribute it and/or modify
    it under the terms of the GNU Lesser Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    BUILTU is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU Lesser Public License for more details.

    You should have received a copy of the GNU Lesser Public License
    along with BUILTU.  If not, see <http://www.gnu.org/licenses/>.
    
    Copyright (C) 2005-2010 Pasko Boris
 */
package org.builtu.util.fs.transactions;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;

import org.apache.log4j.Logger;

/**
 * Simple yet effective transaction mechanism for processing large files. Suppose you have a big file of phone numbers 
 * you want to convert to different format. If you do this as one big step, you always risk that the conversion process 
 * will break down somewhere in the middle and you will lose all data that has been processed up to that point.
 * 
 * What you should do is (program1):
 * <pre>
 * final FileTransaction tran = new FileTransaction("phones.txt", "r");
 * try {
 *   final RandomAccessFile f = tran.getFile();
 *   String line = f.readline();
 *   while (line != null) {
 *     process(line);
 *   }
 *   
 *   tran.commit();
 * } finally {
 *   tran.close();
 * }
 * </pre>
 * 
 * FileTransaction will remember last processed point when you call {@link FileStransaction#commit()} method and will 
 * restart next time from this point on. Therefore, when you ran your program1 again and execute line
 * <pre>
 * final RandomAccessFile f = tran.getFile();
 * </pre>
 * you will get file f already positioned at last committed position.
 *  
 * NOTE: Your file could be binary file, it does not matter.
 * NOTE: You could WRITE to file you process if you specify "rw" or "w" mode
 * NOTE: You are not supposed to create several FileTransaction's on file you process. There is simple rudimentary 
 *       mechanism that tries to prevent you from doing that. But it is not very effective, so if you need to lock your 
 *       file, use {@link FileExclusiveLock}
 * 
 * FileTransaction keeps committed position in file 
 * "phones.txt.FileTransaction.log"
 *  
 * @author borisman
 *
 */
public class FileTransaction {
    private static final Logger logger = Logger.getLogger(FileTransaction.class);
    private final RandomAccessFile file;
    private final File tranLog;
    private long restartPos;

    /**
     * Create FileTransaction object and prepare file at path for processing by seeking it to last committed position
     * @param path File to be processed
     * @param mode access mode of that file, see {@link RandomAccessFile}
     * @throws IOException
     */
    public FileTransaction(final String path, final String mode) throws IOException {
    	this(path, new RandomAccessFile(path, mode));
    }

    /**
     * Create FileTransaction basing on existing acquired lock 
     * @param lock
     * @throws IllegalStateException when lock is not acquired prior to creation of FileTransaction
     */
    public FileTransaction(final FileExclusiveLock lock) throws IOException, IllegalStateException {
	    this(lock.getPath(), makeSureLockObtained(lock));
	}

    private static String makeSureLockObtained(final FileExclusiveLock lock) throws IllegalStateException {
    	if (!lock.isAcquired()) {
    		throw new IllegalStateException("Please acquire lock first"); 
    	}
		return lock.getPath();
	}

    /**
     * Create FileTrasnsaction on already opened file
     * @throws IOException
     */
	public FileTransaction(final String path, final RandomAccessFile file) throws IOException {
	    this.file = file;
	    this.tranLog = new File(path+".FileTransaction.log");
	    this.restartPos = readPos(tranLog);
	    logger.info("Restarting at "+restartPos);
	    file.seek(restartPos);
	}
    
    /**
     * @return file that has to be processed
     */
    public RandomAccessFile getFile() {
        return file;
    }
    
    /**
     * Write current file position to transaction log
     * @throws TransactionFixException when can't update transaction log or concurrent modification detected
     */
    public synchronized void commit() throws TransactionFixException {
        try {
            restartPos = updatePos(tranLog, restartPos, file.getFilePointer());
        } catch (final IOException e) {
            throw new TransactionFixException(e);
        }
    }

    /**
     * Close processed file
     * @throws IOException
     */
    public void close() throws IOException {
        file.close();
    }

    //==========================================================================
    // Implementation 
    //==========================================================================
    
    private static long readPos(final File tranLog) throws IOException {
        final RandomAccessFile f;
        try {
            f = new RandomAccessFile(tranLog, "r");
        } catch (final FileNotFoundException ex) {
            return 0;
        }
        String pos;
        try {
            pos = f.readLine();
        } finally {
            f.close();
        }
        if (pos != null) {
        	pos = pos.trim();
        	if (pos.length()==0) return 0;
        }
        return Long.parseLong(pos);
    }
    
    private static long updatePos(
            final File tranLog, 
            final long oldPos, 
            final long newPos
    ) throws IOException, TransactionFixException {
        final RandomAccessFile f = new RandomAccessFile(tranLog, "rws");
        try {
            final String oldp = f.readLine();
            final long actualOld = oldp == null ? 0 : Long.parseLong(oldp);
            if (oldPos != actualOld) {
            	final String message = "Somebody touched the "+tranLog.getAbsolutePath()+
            	                       " file! Expected to see position="+oldPos+", actual="+oldp;
            	throw new TransactionFixException(message);
            }
                    
            f.seek(0);
            f.setLength(0);
            f.writeBytes(String.valueOf(newPos));
            return newPos;
        } finally {
            f.close();
        }
    }
}
