package org.builtu;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Date;

import javax.sql.DataSource;

import org.apache.log4j.Logger;
import org.builtu.config.module.Config;
import org.builtu.config.module.ConfigUtil;
import org.builtu.config.module.Configurable;


public class JdbcStatListener implements StatListener, Configurable {
	
	private static final Logger log = Logger.getLogger(JdbcStatListener.class);
	public static final String MODULE="webstat-jdbc";

	private String table, insert, create;
	private Connection connection;
	private PreparedStatement stmt;
	
	public JdbcStatListener() {
		ConfigUtil.configure(MODULE, this);
	}
	
	public void close() {
		if (stmt != null) {
			try {
				stmt.close();
			} catch (final SQLException e) {
				log.warn("Error closing statement (non-fatal)", e);
			}
			stmt = null;
		}
		if (connection != null) {
			try {
				connection.close();
			} catch (SQLException e) {
				log.warn("Error closing statement (non-fatal)", e);
			}
			connection = null;
		}
	}
	
	@Override public void configure(String moduleName, Config config) {
		table = config.getString(MODULE+".table");
		insert = config.getString(MODULE+".insert");
		create = config.getString(MODULE+".create-table");
	}


	@Override public void listen(final Date from, final Date to, final int liveSessionsCount) {
		try {
			insert(from, to, liveSessionsCount);
		} catch (final SQLException ex) {
			log.info("Error while inserting record, reconnectiong", ex);
			close();
			try {
				insert(from, to, liveSessionsCount);
			} catch (final SQLException ex2) {
				log.error("Statistics lost", ex2);
			}
		}
		
	}
	
	private void insert(final Date from, final Date to, final int count) throws SQLException {
		final PreparedStatement stmt = getStatement ();
		int i = 1;
		stmt.setDate(i++, new java.sql.Date(from.getTime()));
		stmt.setDate(i++, new java.sql.Date(to.getTime()));
		stmt.setInt(i++, count);
		
		stmt.execute();
	}

	private DataSource dataSource;

	public DataSource getDataSource() {
		return dataSource;
	}

	public void setDataSource(DataSource dataSource) {
		this.dataSource = dataSource;
	}
	
	private PreparedStatement getStatement () throws SQLException {
		if (stmt != null) return stmt;
		final Connection conn = getConnection();
		//if (!tableExists(conn)) createTable(conn);
		try {
			stmt = conn.prepareStatement(insert);
		} catch (final SQLException e) {
			// table may not exists, create one and retry
			final Statement s = conn.createStatement();
			try {
				s.executeUpdate(create);
			} catch (final SQLException ex) {
				log.warn("Can't create table", ex);
			}
			stmt = conn.prepareStatement(insert);
		}
		return stmt;
	}
	
	Connection getConnection() throws SQLException {
		if (connection != null) return connection;
		if (dataSource != null) {
			try {
				connection = dataSource.getConnection();
			} catch (final SQLException e) {
				log.error("Statistics will not be available", e);
				connection = null;
			}
		} else {
			try {
				connection = tryGetConnectionFromConfig();
			} catch (SQLException e) {
				log.error("Statistics will not be available", e);
				connection = null;
			}
		}
		
		if (connection == null) throw new SQLException("Can't get connection");
		
		return connection;
	}
	
	private static class ConnectionGetter implements Configurable {
		
		private String url, username, password;
		
		public ConnectionGetter() {
			ConfigUtil.configure(MODULE, this);
		}

		@Override public void configure(final String moduleName, final Config config) {
			final String driverClassName = config.getString(MODULE+".jdbc.driverClassName");
			url = config.getString(MODULE+".jdbc.url");
			username = config.getString(MODULE+".jdbc.username");
			password = config.getString(MODULE+".jdbc.password");
			
			try {
				Class.forName(driverClassName);
			} catch (final ClassNotFoundException e) {
				log.error("Statistics will not be available", e);
				url = null;
			}
		}
		
		public Connection get () throws SQLException {
			if (url == null) return null;
			return DriverManager.getConnection(url, username, password);
		}
		
	}
	
	private static volatile ConnectionGetter connectionGetter;
	
	private static Connection tryGetConnectionFromConfig() throws SQLException {
		if (connectionGetter!=null) return connectionGetter.get();
		synchronized (JdbcStatListener.class) {
			if (connectionGetter!=null) return connectionGetter.get();
			connectionGetter = new ConnectionGetter();
		}
		return connectionGetter.get();
	}

}
